#############################################################################
##  
#W  lib/Decimal.gi                  FUtil                       Frank Lbeck 
##  
#Y  Copyright (C)  2001
##  
##  This is based on Jean Michel's GAP3 deposit decimal.g. In addition it 
##  implements a  simple interval arithmetic such  that real  cyclotomic 
##  numbers can  be compared strictly.
##  

##  Here are some examples:
##  
##  a1 := DecimalApproximation(3/8);
##  a2 := DecimalApproximation(E(3), 100);
##  a3 := RealPart(a2);
##  a4 := ImaginaryPart(a2);
##  a5 := a4 - a1;
##  a6 := DecimalApproximation(1/2);
##  a7 := RealPart(DecimalApproximation(Sqrt(3)));
##  TruelyLessDecimal(a6,a7);
##  a:=RealPart(DecimalApproximation(Sqrt(3)));
##  for i in [1..20] do a := a*a/a; od;
##  TruelyLessDecimal(a6,a);
##  HasPositiveRealPartCyc(Sqrt(3)-1/2);
##  SqrtDecimalApproximation(3,1000);
##  PiDecimalApproximation(1000);

DefaultDecimalPrecision := 10;

## String for real decimals
NewMethod(String, ["IsDecimalApproximation"],
        function(a)
  local   res,  as,  i;
  res := "";
  if a!.Mantissa < 0 then 
    Add(res,'-');
  fi;
  as := BestQuoInt(AbsInt(a!.Mantissa),10);
  Append(res,String(QuoInt(as,10^(a!.Precision-1))));
  as := String(as mod(10^(a!.Precision-1)));
  as := Concatenation(".",List([1..a!.Precision-1-Length(as)],x->'0'),as);
  i := First([Length(as),Length(as)-1..1],x->as[x]<>'0');
  as := as{[1..i]};
  if as="." then as := "";fi;
  Append(res,as);
  return String(res);
end);

NewMethod(PrintObj, ["IsDecimalApproximation"],
        function(a)
  Print(String(a));
end);


## DecimalApproximation for Rationals
NewMethod(DecimalApproximation, ["IsRat", "IsPosInt"],
        function(x, n)
  local   xx,  res;
  xx := x * 10^n;
  if IsInt(xx) then
    res := rec(Mantissa := xx,
               Precision := n,
               Epsilon := 0);
  else
    res := rec(Mantissa := BestQuoInt(NumeratorRat(xx), DenominatorRat(xx)),
               Precision := n,
               Epsilon := 1);
  fi;
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);    

NewMethod(DecimalApproximation, "IsObject",
        function(a) 
  return DecimalApproximation(a, DefaultDecimalPrecision);
end);

## Changing the precision
NewMethod(DecimalApproximation, ["IsDecimalApproximation", "IsPosInt"],
        function(x, n)
  local   px,  q,  eps,  res;
  px := x!.Precision;
  if n < px then
    q := 10^(px-n);
    eps := x!.Epsilon/q;
    if not IsInt(eps) then
      eps := Int(eps) + 1;
    fi;
    res := rec(Mantissa := BestQuoInt(x!.Mantissa, q), 
               Precision := n,
               Epsilon := eps);
  else
    q := 10^(n-px);
    res := rec(Mantissa := x!.Mantissa * q, 
               Precision := n,
               Epsilon := x!.Epsilon * q);
  fi;
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);
    

## Addition for real decimals
NewMethod(\+, ["IsDecimalApproximation", "IsDecimalApproximation"],
        function(l, r)
  local   pl,  pr,  res;
  pl := l!.Precision;
  pr := r!.Precision;
  if pl < pr then
    r := DecimalApproximation(r, pl);
  elif pr < pl then
    l := DecimalApproximation(l, pr);
  fi;
  res := rec(Mantissa := l!.Mantissa + r!.Mantissa,
             Precision := l!.Precision,
             Epsilon := l!.Epsilon + r!.Epsilon);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

NewMethod(ZeroOp, "IsDecimalApproximation", 
        function(l)
  local   res;
  res := rec(Mantissa := 0,
             Precision := l!.Precision,
             Epsilon := 0);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

NewMethod(AdditiveInverseOp, "IsDecimalApproximation", 
        function(l)
  local   res;
  res := rec(Mantissa := -l!.Mantissa,
             Precision := l!.Precision,
             Epsilon := l!.Epsilon);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

NewMethod(\+, ["IsDecimalApproximation", "IsRat"],
        function(l, r)
  return l + DecimalApproximation(r, l!.Precision);
end);

NewMethod(\+, ["IsRat", "IsDecimalApproximation"],
        function(l, r)
  return DecimalApproximation(l, r!.Precision) + r;
end);

## now the same for multiplication
NewMethod(\*, ["IsDecimalApproximation", "IsDecimalApproximation"],
        function(l,r)
  local   pl,  pr,  pp,  eps,  res;
  pl := l!.Precision;
  pr := r!.Precision;
  if pl < pr then
    r := DecimalApproximation(r, pl);
  elif pr < pl then
    l := DecimalApproximation(l, pr);
    pl := l!.Precision;
  fi;
  pp := 10^pl;
  eps := l!.Epsilon * AbsInt(r!.Mantissa) + 
         r!.Epsilon * AbsInt(l!.Mantissa) + 
         l!.Epsilon * r!.Epsilon;
  if eps <> infinity then
    eps := QuoInt(eps, pp) + 1;
  fi;
  res := rec(Mantissa := BestQuoInt(l!.Mantissa * r!.Mantissa, pp),
             Precision := pl,
             Epsilon := eps);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

NewMethod(OneOp, "IsDecimalApproximation", 
        function(l)
  local   res;
  res := rec(Mantissa := 10^l!.Precision,
             Precision := l!.Precision,
             Epsilon := 0);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

NewMethod(InverseOp, "IsDecimalApproximation", 
        function(l)
  local   m,  q,  eps,  qq,  res;
  m := AbsInt(l!.Mantissa);
  if m = 0 then 
    Error("Division by zero!\n");
  fi;
  q := BestQuoInt(10^(2 * l!.Precision), m);
  if m <= l!.Epsilon then
    eps := infinity;
  else
    qq := BestQuoInt(10^(2 * l!.Precision), m - l!.Epsilon);
    eps := qq - q + 1;
  fi;
  res := rec(Mantissa := SignInt(l!.Mantissa) * q,
             Precision := l!.Precision,
             Epsilon := eps);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

##  to make division more accurate we add this method
NewMethod(\/, ["IsDecimalApproximation", "IsDecimalApproximation"], 
        function(l, r)
  local   pl,  pr,  mr,  ml,  pp,  q,  eps,  qq,  res;
  pl := l!.Precision;
  pr := r!.Precision;
  if pl < pr then
    r := DecimalApproximation(r, pl);
  elif pr < pl then
    l := DecimalApproximation(l, pr);
    pl := l!.Precision;
  fi;
  mr := AbsInt(r!.Mantissa);
  ml := AbsInt(l!.Mantissa);
  if mr = 0 then 
    Error("Division by zero!\n");
  fi;
  pp := 10^pl;
  q := BestQuoInt(ml * pp, mr);
  if mr <= r!.Epsilon or ml <= l!.Epsilon then
    eps := infinity;
  else
    qq := BestQuoInt((ml + l!.Epsilon) * pp, mr - r!.Epsilon);
    eps := qq - q + 1;
  fi;
  res := rec(Mantissa := SignInt(r!.Mantissa) * SignInt(l!.Mantissa) * q,
             Precision := l!.Precision,
             Epsilon := eps);
  SET_TYPE_COMOBJ(res, RealApprType);
  return res;
end);

## now we need arithmetic with infinity
NewMethod(\+, ["IsInfinity", "IsCyc"], 
        function(l, r) 
  return infinity;
end);
NewMethod(\+, ["IsCyc", "IsInfinity"], 
        function(l, r) 
  return infinity;
end);
NewMethod(\+, ["IsInfinity", "IsInfinity"], 
        function(l, r) 
  return infinity;
end);
NewMethod(\*, ["IsInfinity", "IsCyc"], 
        function(l, r) 
  return infinity;
end);
NewMethod(\*, ["IsCyc", "IsInfinity"], 
        function(l, r) 
  return infinity;
end);
NewMethod(\*, ["IsInfinity", "IsInfinity"], 
        function(l, r) 
  return infinity;
end);

## multiplication with Rationals
NewMethod(\*, ["IsDecimalApproximation", "IsRat"],
        function(l, r)
  return l * DecimalApproximation(r, l!.Precision);
end);

NewMethod(\*, ["IsRat", "IsDecimalApproximation"],
        function(l, r)
  return DecimalApproximation(l, r!.Precision) * r;
end);

##########################################################################
##  Now the arithmetic for complex numbers a + b I

## printing of complex numbers: a+I*b
NewMethod(String, "IsComplexNumber",
        function(l)
  local   sr,  si,  sig;
  sr := String(l!.RealPart);
  si := String(l!.ImaginaryPart);
  if sr = "0" then
    if si = "0" then
      return si;
    elif si = "1" then
      return "I";
    elif si = "-1" then
      return "-I";
    else
      if si[1] = '-' then
        sig := "-";
        si := si{[2..Length(si)]};
      else
        sig := "";
      fi;
      if '+' in si or '-' in si then
        si := Concatenation("(", si, ")");
      fi;
      return Concatenation(sig, "I*", si);
    fi;
  else
    if si = "0" then
      return sr;
    elif si = "1" then
      return Concatenation(sr, "+I");
    elif si = "-1" then
      return Concatenation(sr, "-I");
    else
      if si[1] = '-' then
        sig := "-";
        si := si{[2..Length(si)]};
      else
        sig := "+";
      fi;
      if '+' in si or '-' in si then
        si := Concatenation("(", si, ")");
      fi;
      return Concatenation(sr, sig, "I*", si);
    fi;
  fi;
end);

NewMethod(PrintObj, "IsComplexNumber", 
        function(l)
  Print(String(l));
end);

NewMethod(ComplexNumber, "IsCyc",
        function(l)
  local   lb,  res;
  lb := GaloisCyc(l, -1);
  res := rec(RealPart := (l + lb) / 2,
             ImaginaryPart := (l - lb) / 2 * E(4)^3);
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);
NewMethod(ComplexNumber, ["IsRingElement", "IsRingElement"],
        function(re, im)
  local   res;
  res := rec(RealPart := re,
             ImaginaryPart := im);
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);
# default is to consider l as real part
NewMethod(ComplexNumber, "IsObject",  function(l)
  return ComplexNumber(l, 0*l);
end);


NewMethod(DecimalApproximation, ["IsComplexNumber", "IsPosInt"],
        function(l, n)
  local   res;
  res := rec(RealPart := DecimalApproximation(l!.RealPart, n),
             ImaginaryPart := DecimalApproximation(l!.ImaginaryPart, n));
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);
  
##  additive structure
NewMethod(\+, ["IsComplexNumber", "IsComplexNumber"],
        function(l, r)
  local   res;
  res := rec(RealPart := l!.RealPart + r!.RealPart,
             ImaginaryPart := r!.ImaginaryPart + l!.ImaginaryPart);
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);

NewMethod(ZeroOp, "IsComplexNumber",
        function(l)
  local   res;
  res := rec(RealPart := ZeroOp(l!.RealPart),
             ImaginaryPart := ZeroOp(l!.ImaginaryPart));
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);
  
NewMethod(AdditiveInverseOp, "IsComplexNumber",
        function(l)
  local   res;
  res := rec(RealPart := AdditiveInverseOp(l!.RealPart),
             ImaginaryPart := AdditiveInverseOp(l!.ImaginaryPart));
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);
  
NewMethod(\+, ["IsComplexNumber", "IsObject"],
        function(l, r)
  return l + ComplexNumber(r);
end);
NewMethod(\+, ["IsObject", "IsComplexNumber"],
        function(l, r)
  return ComplexNumber(l) + r;
end);

##  multiplicative structure
NewMethod(\*, ["IsComplexNumber", "IsComplexNumber"],
        function(l, r)
  local   res;
  res := rec(RealPart := l!.RealPart * r!.RealPart - l!.ImaginaryPart
             * r!.ImaginaryPart,
             ImaginaryPart := l!.RealPart * r!.ImaginaryPart +
             l!.ImaginaryPart * r!.RealPart);
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);

NewMethod(OneOp, "IsComplexNumber",
        function(l)
  local   res;
  res := rec(RealPart := OneOp(l!.RealPart),
             ImaginaryPart := ZeroOp(l!.ImaginaryPart));
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);

NewMethod(InverseOp, "IsComplexNumber",
        function(l)
  local   n,  res;
  n := l!.RealPart^2 + l!.ImaginaryPart^2;
  res := rec(RealPart := l!.RealPart / n,
             ImaginaryPart := l!.ImaginaryPart / n);
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);

NewMethod(\*, ["IsComplexNumber", "IsObject"],
        function(l, r)
  return l * ComplexNumber(r);
end);
NewMethod(\*, ["IsObject", "IsComplexNumber"],
        function(l, r)
  return ComplexNumber(l) * r;
end);

NewMethod(RealPart, "IsDecimalApproximation", IdFunc);
  
######################################################################
## first n + 1 summands of exponential series
NewMethod(Exp, [IsObject, IsInt], function(x, n)
  local res, y, i;
  res := OneOp(x);
  y := res;
  for i in [1..n] do
    y := (1/i) * y * x ;
    res := res + y;  
  od;
  return res;
end);
######################################################################
## Exp for nilpotent elements
NewMethod(Exp, [IsObject], function(x)
  local res, y, zero, i;
  res := OneOp(x);
  y := res;
  zero := ZeroOp(y);
  i := 1;
  while y <> zero do
    y := y * x / i;
    res := res + y;
    i := i + 1;
  od;
  return res;
end);


InstallGlobalFunction(NewtonRoU, function(x, n)
  return ((n-1) * x + 1 / x^(n-1)) / n; 
end);

InstallGlobalFunction(NewtonSqrt, function(x, a)
  return (x * x + a) / (x + x);
end);

##  sqrt by Newton approximation
InstallGlobalFunction(SqrtDecimalApproximation, function(a, n)
  local   x,  y,  diff;
  a := DecimalApproximation(a, n);
  if n > 10 then
    x := DecimalApproximation(SqrtDecimalApproximation(a, QuoInt(n, 2)), n);
  else
    x := a;
  fi;
  y := ZeroOp(x);
  if IsDecimalApproximation(a) then 
    diff := AbsInt(y!.Mantissa - x!.Mantissa);
    while diff > 9 do
      y := x;
      x := NewtonSqrt(y, a);
      diff := AbsInt(y!.Mantissa - x!.Mantissa);
    od;
    x!.Epsilon := diff + 1;
  elif IsComplexNumber(x) then
    diff := Maximum(AbsInt(y!.RealPart!.Mantissa - x!.RealPart!.Mantissa),
                    AbsInt(y!.ImaginaryPart!.Mantissa - 
                           x!.ImaginaryPart!.Mantissa));
    while diff > 9 do
      y := x;
      x := NewtonSqrt(y, a);
      diff := Maximum(AbsInt(y!.RealPart!.Mantissa - x!.RealPart!.Mantissa),
                      AbsInt(y!.ImaginaryPart!.Mantissa - 
                             x!.ImaginaryPart!.Mantissa));
    od;
    x!.RealPart!.Epsilon := diff + 1;
    x!.ImaginaryPart!.Epsilon := diff + 1;
  else
    Error("non valid argument: ", a, "\n");
  fi;
  return x;
end);

##  recursive approximation of pi with exponential convergence, see
##  Koecher, Klassische elementare Analysis, Birkhuser, 1987
PIDECIMAL := DecimalApproximation(22/7, 2);
BindGlobal("PiDecimalApproximation", function(n)
  local   alpha,  beta,  pi,  lastpi,  diff,  sqrtalpha;
  if n <= PIDECIMAL!.Precision then
    return DecimalApproximation(PIDECIMAL, n);
  fi;
  alpha := SqrtDecimalApproximation(2, n);
  beta := ZeroOp(alpha);
  pi := 2 + alpha;
  lastpi := beta;
  diff := AbsInt(pi!.Mantissa - lastpi!.Mantissa);
  while diff > 9 do
    lastpi := pi;
    sqrtalpha := SqrtDecimalApproximation(alpha, n);
    beta := sqrtalpha * (beta + 1) / (beta + alpha);
    alpha := (sqrtalpha + 1 / sqrtalpha) / 2;
    pi := pi * beta * (1 + alpha) / (1 + beta);
    diff := AbsInt(pi!.Mantissa - lastpi!.Mantissa);
  od;
  pi!.Epsilon := diff + 1;
  PIDECIMAL := pi;
  return pi;
end);  

## we cache real approximations of E(i)^r
ROUREALDECIMAL := [];
BindGlobal("RealPartCycDecimalApproximation", function(l, n)
  local   r,  pi,  res,  i, I;
  # this makes sure that we detect 0 exactly
  l := ExtRepOfObj((l + GaloisCyc(l, -1)) / 2);
  r := Length(l);
  if not IsBound(ROUREALDECIMAL[r]) or 
     ROUREALDECIMAL[r][1]!.Precision < n then
    pi := PiDecimalApproximation(n);
    ROUREALDECIMAL[r] := [DecimalApproximation(1, n)]; 
    I := ComplexNumber(0, ROUREALDECIMAL[r][1]);
    ROUREALDECIMAL[r][r+1] := Exp(2*pi*I / r, n+5);
  fi;
  res := DecimalApproximation(0, n);
  for i in [1..r] do
    if l[i] <> 0 then
      if not IsBound(ROUREALDECIMAL[r][i]) then
        ROUREALDECIMAL[r][i] := RealPart(ROUREALDECIMAL[r][r+1]^(i - 1));
      fi;
      res := res + l[i] * DecimalApproximation(ROUREALDECIMAL[r][i], n);
    fi;
  od;
  return res;
end);

NewMethod(DecimalApproximation, ["IsCyc", "IsPosInt"],
        function(l, n)
  local   res;
  # real case
  if GaloisCyc(l, -1) = l then
    return RealPartCycDecimalApproximation(l, n);
  fi;
  res := rec( RealPart := RealPartCycDecimalApproximation(l, n),
              ImaginaryPart :=
              RealPartCycDecimalApproximation(-E(4)*l, n));
  SET_TYPE_COMOBJ(res, ComplexType);
  return res;
end);

##  for convenience to apply DecimalApproximation to all entries in
##  a list or matrices or ...
NewMethod(DecimalApproximation, ["IsList", "IsPosInt"],
        function(l, n)
  return List(l, a-> DecimalApproximation(a, n));
end);

##  \< for decimal real numbers, gives `fail' when l < r but 
##  l + Epsilon(l) >= r - Epsilon(r)  (or the other way round)
BindGlobal("TruelyLessDecimal", function(l, r)
  local   pl,  pr;
  if IsInfinity(r!.Epsilon) or IsInfinity(l!.Epsilon) then
    return fail;
  fi;
  pl := l!.Precision;
  pr := r!.Precision;
  if pl < pr then
    r := DecimalApproximation(r, pl);
  elif pr < pl then
    l := DecimalApproximation(l, pr);
    pl := l!.Precision;
  fi;
  if l!.Mantissa + l!.Epsilon < r!.Mantissa - r!.Epsilon then
    return true;
  elif l!.Mantissa - l!.Epsilon >= r!.Mantissa + r!.Epsilon then
    return false;
  else
    return fail;
  fi;
end);

BindGlobal("HasPositiveRealPartCyc", function(x)
  local   pre,  a;
  if IsRat(x) then 
    return x > 0;
  fi;
  pre := 6;
  a := RealPartCycDecimalApproximation(x, pre);
  if a!.Mantissa = 0 and a!.Epsilon = 0 then
    return false;
  fi;
  while a!.Epsilon >= AbsInt(a!.Mantissa) do
    pre := 3 * pre;
    a := RealPartCycDecimalApproximation(x, pre);
  od;
  return IsPosInt(a!.Mantissa);
end);

